
from enum import IntEnum
from NooLite_F.Adapter import Adapter, Response, Request, Command, Mode, ResponseCode, Action


class ModuleState(IntEnum):
    OFF = 0,
    ON = 1,
    TEMPORARY_ON = 2


class ModuleMode(IntEnum):
    BIND_OFF = 0,
    BIND_ON = 1,


class BrightnessDirection(IntEnum):
    UP = 0,
    DOWN = 1,


class ModuleInfo(object):
    state: ModuleState
    mode: ModuleMode
    brightness: float = None
    id: int = None
    firmware: int = None
    type: int = None

    def __repr__(self):
        return "<ModuleInfo (0x{0:x}), id: 0x{1:x}, type: {2}, firmware: {3}, state: {4}, brightness: {5}, mode: {6}>"\
            .format(id(self), self.id, self.type, self.firmware, self.state, self.brightness, self.mode)


class ModuleInfoParser(object):
    @staticmethod
    def parse(response: Response) -> ModuleInfo:
        info = None
        if response.command == Command.SEND_STATE:
            info = ModuleInfo()
            info.type = response.data[0]
            info.firmware = response.data[1]
            info.id = response.id

            if response.format == 0:
                info.state = ModuleState(response.data[2] & 0x0F)
                info.mode = ModuleMode(response.data[2] & 0x80)
                info.brightness = response.data[3] / 255

        return info


# TODO: replace format with enum or constants???
class NooLiteService(object):

    _adapter = None

    def __init__(self, port: str):
        self._adapter = Adapter(port)

    def off(self, channel: int, broadcast: bool = False, mode: Mode = Mode.TX_F) -> [(bool, ModuleInfo)]:
        """ Turn off the modules

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        responses = self._send_module_command(channel, Command.OFF, broadcast, mode)
        return self._handle_command_responses(responses)

    def on(self, channel: int, broadcast: bool = False, mode: Mode = Mode.TX_F) -> [(bool, ModuleInfo)]:
        """ Turn on the modules

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        responses = self._send_module_command(channel, Command.ON, broadcast, mode)
        return self._handle_command_responses(responses)

    def temporary_on(self, channel: int, duration: int, broadcast: bool = False, mode: Mode = Mode.TX_F) -> [(bool, ModuleInfo)]:
        """ Turn on the modules for a specified time interval

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param duration: the time during which the modules will be turned on, duration measurement equals 5 sec.
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        data: bytearray = bytearray(2)
        data[0] = duration & 0x00FF
        data[1] = duration & 0xFF00

        responses = self._send_module_command(channel, Command.TEMPORARY_ON, broadcast, mode, data, 6)
        return self._handle_command_responses(responses)

    def enable_temporary_on(self, channel: int, broadcast: bool = False, mode: Mode = Mode.TX_F) -> [(bool, ModuleInfo)]:
        """ Enable "temporary on" mode

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        data: bytearray = bytearray(1)
        data[0] = 0

        responses = self._send_module_command(channel, Command.MODES, broadcast, mode, data, 1)
        return self._handle_command_responses(responses)

    def disable_temporary_on(self, channel: int, broadcast: bool = False, mode: Mode = Mode.TX_F) -> [(bool, ModuleInfo)]:
        """ Disable "temporary on" mode

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        data: bytearray = bytearray(1)
        data[0] = 1

        responses = self._send_module_command(channel, Command.MODES, broadcast, mode, data, 1)
        return self._handle_command_responses(responses)

    def switch(self, channel: int, broadcast: bool = False, mode: Mode = Mode.TX_F) -> [(bool, ModuleInfo)]:
        """ Switch modules mode (on/off)

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        responses = self._send_module_command(channel, Command.SWITCH, broadcast, mode)
        return self._handle_command_responses(responses)

    def brightness_tune(self, channel: int, direction: BrightnessDirection, broadcast: bool = False, mode: Mode = Mode.TX_F) -> [(bool, ModuleInfo)]:
        """ Start to increase/decrease brightness

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param direction: direction of the brightness changing
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        if direction == BrightnessDirection.UP:
            command = Command.BRIGHT_UP
        else:
            command = Command.BRIGHT_DOWN

        responses = self._send_module_command(channel, command, broadcast, mode)
        return self._handle_command_responses(responses)

    def brightness_tune_back(self, channel: int, broadcast: bool = False, mode: Mode = Mode.TX_F) -> [(bool, ModuleInfo)]:
        """ Invert direction of the brightness change

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        responses = self._send_module_command(channel, Command.BRIGHT_BACK, broadcast, mode)
        return self._handle_command_responses(responses)

    def brightness_tune_stop(self, channel: int, broadcast: bool = False, mode: Mode = Mode.TX_F) -> [(bool, ModuleInfo)]:
        """ Stop brightness changing

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        responses = self._send_module_command(channel, Command.STOP_BRIGHT, broadcast, mode)
        return self._handle_command_responses(responses)

    def brightness_tune_custom(self, channel: int, direction: BrightnessDirection, speed: float, broadcast: bool = False, mode: Mode = Mode.TX_F):
        """ Start to increase/decrease brightness with a specified speed

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param direction: direction of the brightness changing
        :param speed: speed of the brightness changing. The range of value is 0 .. 1.0
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """

        if speed >= 1:
            value = 127
        elif speed <= 0:
            value = 0
        else:
            value = int(speed * 127)

        if direction == BrightnessDirection.DOWN:
            value = -value - 1

        data: bytearray = bytearray(1)
        data[0] = value & 0xFF

        responses = self._send_module_command(channel, Command.BRIGHT_REG, broadcast, mode, data, 1)
        return self._handle_command_responses(responses)

    def brightness_tune_step(self, channel: int, direction: BrightnessDirection, step: int = None, broadcast: bool = False, mode: Mode = Mode.TX_F):
        """ Increase/decrease brightness once with a specified step

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param direction: direction of the brightness changing
        :param step: step in microseconds. If specify then can have values in range (1..255) or 0 (it is means 256), by default step equals 64
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        data: bytearray = None
        fmt: int = None

        if step is not None:
            fmt = 1
            data: bytearray = bytearray(1)
            data[0] = step

        if direction == BrightnessDirection.UP:
            command = Command.BRIGHT_STEP_UP
        else:
            command = Command.BRIGHT_STEP_DOWN

        responses = self._send_module_command(channel, command, broadcast, mode, data, fmt)
        return self._handle_command_responses(responses)

    def set_brightness(self, channel: int, brightness: float, broadcast: bool = False, mode: Mode = Mode.TX_F) -> [(bool, ModuleInfo)]:
        """ Set brightness

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param brightness: brightness level. The range of value is 0 .. 1.0
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        if brightness >= 1:
            value = 255
        elif brightness <= 0:
            value = 0
        else:
            value = 35 + int(120 * brightness)

        print(value)

        data: bytearray = bytearray(1)
        data[0] = value

        responses = self._send_module_command(channel, Command.SET_BRIGHTNESS, broadcast, mode, data, 1)
        return self._handle_command_responses(responses)

    def roll_rgb_color(self, channel: int, broadcast: bool = False, mode: Mode = Mode.TX_F) -> [(bool, ModuleInfo)]:
        """ Start color changing (only for RGB Led modules)

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        responses = self._send_module_command(channel, Command.ROLL_COLOR, broadcast, mode)
        return self._handle_command_responses(responses)

    def switch_rgb_color(self, channel: int, broadcast: bool = False, mode: Mode = Mode.TX_F) -> [(bool, ModuleInfo)]:
        """ Switch color (only for RGB Led modules)

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        responses = self._send_module_command(channel, Command.SWITCH_COLOR, broadcast, mode)
        return self._handle_command_responses(responses)

    def switch_rgb_mode(self, channel: int, broadcast: bool = False, mode: Mode = Mode.TX_F) -> [(bool, ModuleInfo)]:
        """ Switch color changing modes (only for RGB Led modules)

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        responses = self._send_module_command(channel, Command.SWITCH_MODE, broadcast, mode)
        return self._handle_command_responses(responses)

    def switch_rgb_mode_speed(self, channel: int, broadcast: bool = False, mode: Mode = Mode.TX_F) -> [(bool, ModuleInfo)]:
        """ Switch speed of the color changing (only for RGB Led modules)

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        responses = self._send_module_command(channel, Command.SPEED_MODE, broadcast, mode)
        return self._handle_command_responses(responses)

    def set_rgb_brightness(self, channel: int, red: float, green: float, blue: float, broadcast: bool = False, mode: Mode = Mode.TX_F) -> [(bool, ModuleInfo)]:
        """ Set brightness for each rgb color (only for RGB Led modules)

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param red: red color brightness level. The range of value is 0 .. 1.0
        :param green: green color brightness level. The range of value is 0 .. 1.0
        :param blue: blue color brightness level. The range of value is 0 .. 1.0
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        data: bytearray = bytearray(3)
        data[0] = self._convert_color_brightness(red)
        data[1] = self._convert_color_brightness(green)
        data[2] = self._convert_color_brightness(blue)

        responses = self._send_module_command(channel, Command.SET_BRIGHTNESS, broadcast, mode, data, 3)
        return self._handle_command_responses(responses)

    def load_preset(self, channel: int, broadcast: bool = False, mode: Mode = Mode.TX_F) -> [(bool, ModuleInfo)]:
        """ Load saved module state from preset

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        responses = self._send_module_command(channel, Command.LOAD_PRESET, broadcast, mode)
        return self._handle_command_responses(responses)

    def save_preset(self, channel: int, broadcast: bool = False, mode: Mode = Mode.TX_F) -> [(bool, ModuleInfo)]:
        """ Save current module state as preset

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        responses = self._send_module_command(channel, Command.SAVE_PRESET, broadcast, mode)
        return self._handle_command_responses(responses)

    def read_state(self, channel: int, broadcast: bool = False, mode: Mode = Mode.TX_F) -> [(bool, ModuleInfo)]:
        """  Read module state (only for NooLite-F modules)

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        responses = self._send_module_command(channel, Command.READ_STATE, broadcast, mode)
        return self._handle_command_responses(responses)

    def bind(self, channel: int, broadcast: bool = False, mode: Mode = Mode.TX_F) -> [(bool, ModuleInfo)]:
        """  Send bind command to module

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        responses = self._send_module_command(channel, Command.BIND, broadcast, mode)
        return self._handle_command_responses(responses)

    def unbind(self, channel: int, broadcast: bool = False, mode: Mode = Mode.TX_F) -> [(bool, ModuleInfo)]:
        """  Send unbind command to module

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        responses = self._send_module_command(channel, Command.UNBIND, broadcast, mode)
        return self._handle_command_responses(responses)

    def service_mode_on(self, channel: int, broadcast: bool = False, mode: Mode = Mode.TX_F) -> [(bool, ModuleInfo)]:
        """  Turn on the service mode on module (only for NooLite-F modules)

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        data: bytearray = bytearray(1)
        data[0] = 1

        responses = self._send_module_command(channel, Command.SERVICE, broadcast, mode, data)
        return self._handle_command_responses(responses)

    def service_mode_off(self, channel: int, broadcast: bool = False, mode: Mode = Mode.TX_F) -> [(bool, ModuleInfo)]:
        """  Turn off the service mode on module (only for NooLite-F modules)

        :param channel: the number of the channel for command. The command will be send to all modules that are binded with selected channel.
        :param broadcast: broadcast mode for command. If True then command will be send simultaneously to all modules that are binded with selected channel (default - False)
        :param mode: mode of the command sending. TX - for nooLite module (without feedback), TX_F - for noolite-F modules (with feedback).
        :return: for nooLite-F command returns array which contains command result and module info for each module that are binded with selected channel. For nooLite modules returns nothing.
        """
        responses = self._send_module_command(channel, Command.SERVICE, broadcast, mode)
        return self._handle_command_responses(responses)

    def _send_module_command(self, channel: int, command: Command, broadcast, mode: Mode, data: bytearray = None, fmt: int = None) -> [Response]:
        request = Request()

        request.mode = mode
        request.channel = channel
        request.command = command
        if broadcast:
            request.action = Action.SEND_BROADCAST_COMMAND
        else:
            request.action = Action.SEND_COMMAND

        if data is not None:
            request.data = data

        if fmt is not None:
            request.format = fmt

        responses = self._send_request(request)
        return responses

    def _send_request(self, request: Request) -> [Response]:
        self._adapter.open()
        responses = self._adapter.send_request(request)
        self._adapter.close()
        return responses

    @staticmethod
    def _handle_command_responses(responses) -> [(bool, ModuleInfo)]:
        results = []
        for response in responses:
            info = ModuleInfoParser.parse(response)
            status = response.status == ResponseCode.SUCCESS or response.status == ResponseCode.BIND_SUCCESS
            results.append((status, info))

        return results

    @staticmethod
    def _convert_color_brightness(bright: float) -> int:
        if bright >= 1:
            value = 255
        elif bright <= 0:
            value = 0
        else:
            value = int(255 * bright)

        return value




from enum import IntEnum
from serial import Serial


class Command(IntEnum):
    OFF = 0,
    BRIGHT_DOWN = 1,
    ON = 2,
    BRIGHT_UP = 3,
    SWITCH = 4,
    BRIGHT_BACK = 5,
    SET_BRIGHTNESS = 6,
    LOAD_PRESET = 7,
    SAVE_PRESET = 8,
    UNBIND = 9,
    STOP_BRIGHT = 10,
    BRIGHT_STEP_DOWN = 11,
    BRIGHT_STEP_UP = 12,
    BRIGHT_REG = 13,
    BIND = 15,
    ROLL_COLOR = 16,
    SWITCH_COLOR = 17,
    SWITCH_MODE = 18,
    SPEED_MODE = 19,
    BATTERY_LOW = 20,
    SENS_TEMP_HUMI = 21,
    TEMPORARY_ON = 25,
    MODES = 26,
    READ_STATE = 128,
    WRITE_STATE = 129,
    SEND_STATE = 130,
    SERVICE = 131,
    CLEAR_MEMORY = 132


class Mode(IntEnum):
    TX = 0,
    RX = 1,
    TX_F = 2,
    RX_F = 3,
    SERVICE = 4,
    FIRMWARE_UPDATE = 5


class ResponseCode(IntEnum):
    SUCCESS = 0,
    NO_RESPONSE = 1,
    ERROR = 2,
    BIND_SUCCESS = 3


class Action(IntEnum):
    SEND_COMMAND = 0,
    SEND_BROADCAST_COMMAND = 1,
    READ_RESPONSE = 2,
    BIND_MODE_ON = 3,
    BIND_MODE_OFF = 4,
    CLEAR_CHANNEL = 5,
    CLEAR_MEMORY = 6,
    UNBIND_ADDRESS_FROM_CHANNEL = 7,
    SEND_COMMAND_BY_ADDRESS = 8


class Request(object):
    mode: Mode = Mode.TX_F
    action: Action = Action.SEND_COMMAND
    channel: int = 0
    command: Command = Command.OFF
    format: int = 0
    data: bytearray = bytearray()
    id: int = 0

    def __repr__(self):
        return "<Request (0x{0:x}), mode: {1}, action: {2}, channel: {3:d}, command: {4:d}, format: {5:d}, data: {6}, id: 0x{7:x}>"\
            .format(id(self), self.mode, self.action, self.channel, self.command, self.format, self.data, self.id)


class Response(object):
    mode: Mode = None
    status: ResponseCode = None
    channel: int = None
    command: Command = None
    count: int = None
    format: int = None
    data: bytearray = None
    id: int = None

    def __repr__(self):
        return "<Response (0x{0:x}), mode: {1}, status: {2}, channel: {3:d}, command: {4:d}, format: {5:d}, data: {6}, id: 0x{7:x}>"\
            .format(id(self), self.mode, self.status, self.channel, self.command, self.format, self.data, self.id)


class PacketUtils:
    @staticmethod
    def crc(data):
        sum = 0
        for i in range(0, len(data) - 2):
            sum = sum + data[i]
        sum = sum & 0xFF
        return sum


class RequestPacketBuilder:
    START_BYTE = 171
    STOP_BYTE = 172

    PACKET_SIZE = 17

    ST = 0
    MODE = 1
    CTR = 2
    CH = 4
    CMD = 5
    FMT = 6
    D0 = 7
    D1 = 8
    D2 = 9
    D3 = 10
    ID0 = 11
    ID1 = 12
    ID2 = 13
    ID3 = 14
    CRC = 15
    SP = 16

    @staticmethod
    def build(request: Request) -> bytearray:
        data = bytearray(RequestPacketBuilder.PACKET_SIZE)

        data[RequestPacketBuilder.ST] = RequestPacketBuilder.START_BYTE
        data[RequestPacketBuilder.MODE] = request.mode.value
        data[RequestPacketBuilder.CTR] = request.action.value
        data[RequestPacketBuilder.CH] = request.channel
        data[RequestPacketBuilder.CMD] = request.command.value
        data[RequestPacketBuilder.FMT] = request.format
        if len(request.data) >= 1:
            data[RequestPacketBuilder.D0] = request.data[0]
        if len(request.data) >= 2:
            data[RequestPacketBuilder.D1] = request.data[1]
        if len(request.data) >= 3:
            data[RequestPacketBuilder.D2] = request.data[2]
        if len(request.data) >= 4:
            data[RequestPacketBuilder.D3] = request.data[3]
        data[RequestPacketBuilder.ID0] = request.id & 0xFF000000
        data[RequestPacketBuilder.ID1] = request.id & 0x00FF0000
        data[RequestPacketBuilder.ID2] = request.id & 0x0000FF00
        data[RequestPacketBuilder.ID3] = request.id & 0x000000FF
        data[RequestPacketBuilder.CRC] = PacketUtils.crc(data)
        data[RequestPacketBuilder.SP] = RequestPacketBuilder.STOP_BYTE

        return data


class ResponseException(Exception):
    """Base class for response exceptions."""


class ResponsePacketParser:
    START_BYTE = 173
    STOP_BYTE = 174

    PACKET_SIZE = 17

    ST = 0
    MODE = 1
    CTR = 2
    TOGL = 3
    CH = 4
    CMD = 5
    FMT = 6
    D0 = 7
    D1 = 8
    D2 = 9
    D3 = 10
    ID0 = 11
    ID1 = 12
    ID2 = 13
    ID3 = 14
    CRC = 15
    SP = 16

    @staticmethod
    def parse(data: bytearray) -> Response:
        if (len(data) != ResponsePacketParser.PACKET_SIZE or data[ResponsePacketParser.ST] != ResponsePacketParser.START_BYTE) or (data[ResponsePacketParser.SP] != ResponsePacketParser.STOP_BYTE) or (data[ResponsePacketParser.CRC] != PacketUtils.crc(data)):
            raise ResponseException("Invalid response")

        response = Response()

        response.mode = Mode(data[ResponsePacketParser.MODE])
        response.status = ResponseCode(data[ResponsePacketParser.CTR])
        response.command = Command(data[ResponsePacketParser.CMD])
        response.channel = data[ResponsePacketParser.CH]
        response.count = data[ResponsePacketParser.TOGL]
        response.format = data[ResponsePacketParser.FMT]
        response.data = bytearray(4)
        response.data[0] = data[ResponsePacketParser.D0]
        response.data[1] = data[ResponsePacketParser.D1]
        response.data[2] = data[ResponsePacketParser.D2]
        response.data[3] = data[ResponsePacketParser.D3]
        response.id = (data[ResponsePacketParser.ID0] << 24) + (data[ResponsePacketParser.ID1] << 16) + (data[ResponsePacketParser.ID2] << 8) + data[ResponsePacketParser.ID3]

        return response


class Adapter(object):
    _port = None
    _serial = Serial()

    def __init__(self, port: str):
        self._port = port

    def open(self):
        self._serial.port = self._port
        self._serial.baudrate = 9600
        self._serial.timeout = 3
        self._serial.open()

    def close(self):
        self._serial.close()

    def read_response(self) -> Response:
        data = self._serial.read(ResponsePacketParser.PACKET_SIZE)
        response = ResponsePacketParser.parse(data)
        print("Receive:\n - packet: {0},\n - response: {1}".format(data, response))
        return response

    def send_request(self, request: Request):
        responses = []

        data = RequestPacketBuilder.build(request)
        print("Send:\n - request: {0},\n - packet: {1}".format(request, data))

        try:
            self._serial.write(data)

            response = self.read_response()
            responses.append(response)

            while response.count > 0:
                response = self.read_response()
                responses.append(response)

        except ResponseException:
            raise

        return responses





from NooLite_F.Modules import *

# adapter = Adapter("COM3")
#
# adapter.open()
#
# while True:
#    response = adapter.read_response()
#    print(response)
#
# adapter.close()


#noolite = NooLiteService("COM3")

#noolite.bind(62, mode=Mode.TX)

#noolite.on(62, mode=Mode.TX)
#noolite.off(62, mode=Mode.TX)

#noolite.bright_tune(62, BrightnessDirection.DOWN, mode=Mode.TX)
#noolite.bright_tune_back(62, mode=Mode.TX)
#noolite.bright_tune_custom(62, BrightnessDirection.UP, 1.0, mode=Mode.TX)
#noolite.bright_tune_stop(62, mode=Mode.TX)
#noolite.bright_step(62, BrightnessDirection.DOWN, 0, mode=Mode.TX)

#noolite.set_rgb_brightness(62, 0.5, 0.5, 0.5, mode=Mode.TX)

#noolite.roll_rgb_color(62, mode=Mode.TX)
#noolite.switch_rgb_color(62, mode=Mode.TX)
#noolite.switch_rgb_mode(62, mode=Mode.TX)
#noolite.switch_rgb_mode_speed(62, mode=Mode.TX)


rgb = RGBLed("COM3", 62, ModuleType.NOOLITE)

#rgb.on()
#rgb.set_rgb_brightness(1.0, 0.0, 0.0)
rgb.off()



from enum import IntEnum

from NooLite_F.NooLiteService import NooLiteService, ModuleInfo, Mode, BrightnessDirection


class ModuleType(IntEnum):
    NOOLITE = 0
    NOOLITE_F = 1


class Switch(object):
    _port: str
    _channel: int
    _module_type: ModuleType
    _broadcast_mode: bool

    _adapter: NooLiteService

    _mode_map = {
        ModuleType.NOOLITE: Mode.TX,
        ModuleType.NOOLITE_F: Mode.TX_F,
    }

    def __init__(self, port: str, channel: int, module_type: ModuleType, broadcast_mode: bool = False):
        self._channel = channel
        self._module_type = module_type
        self._adapter = NooLiteService(port)
        self._broadcast_mode = broadcast_mode

    @property
    def _mode(self) -> Mode:
        return self._mode_map[self._module_type]

    def on(self) -> [(bool, ModuleInfo)]:
        return self._adapter.on(self._channel, self._broadcast_mode, self._mode)

    def off(self) -> [(bool, ModuleInfo)]:
        return self._adapter.off(self._channel, self._broadcast_mode, self._mode)

    def switch(self) -> [(bool, ModuleInfo)]:
        return self._adapter.off(self._channel, self._broadcast_mode, self._mode)

    def temporary_on(self, duration: int) -> [(bool, ModuleInfo)]:
        return self._adapter.temporary_on(self._channel, duration, self._broadcast_mode, self._mode)

    def enable_temporary_on(self) -> [(bool, ModuleInfo)]:
        return self._adapter.enable_temporary_on(self._channel, self._broadcast_mode, self._mode)

    def disable_temporary_on(self) -> [(bool, ModuleInfo)]:
        return self._adapter.disable_temporary_on(self._channel, self._broadcast_mode, self._mode)

    def load_preset(self) -> [(bool, ModuleInfo)]:
        return self._adapter.load_preset(self._channel, self._broadcast_mode, self._mode)

    def save_preset(self) -> [(bool, ModuleInfo)]:
        return self._adapter.save_preset(self._channel, self._broadcast_mode, self._mode)

    def read_state(self) -> [(bool, ModuleInfo)]:
        return self._adapter.read_state(self._channel, self._broadcast_mode, self._mode)

    def bind(self) -> [(bool, ModuleInfo)]:
        return self._adapter.bind(self._channel, self._broadcast_mode, self._mode)

    def unbind(self) -> [(bool, ModuleInfo)]:
        return self._adapter.bind(self._channel, self._broadcast_mode, self._mode)

    def service_mode_on(self) -> [(bool, ModuleInfo)]:
        return self._adapter.service_mode_on(self._channel, self._broadcast_mode, self._mode)

    def service_mode_off(self) -> [(bool, ModuleInfo)]:
        return self._adapter.service_mode_off(self._channel, self._broadcast_mode, self._mode)


class Dimmer(Switch):

    def brightness_tune(self, direction: BrightnessDirection) -> [(bool, ModuleInfo)]:
        return self._adapter.brightness_tune(self._channel, direction, self._broadcast_mode, self._mode)

    def brightness_tune_back(self) -> [(bool, ModuleInfo)]:
        return self._adapter.brightness_tune_back(self._channel, self._broadcast_mode, self._mode)

    def brightness_tune_stop(self) -> [(bool, ModuleInfo)]:
        return self._adapter.brightness_tune_stop(self._channel, self._broadcast_mode, self._mode)

    def brightness_tune_custom(self, direction: BrightnessDirection, speed: float):
        return self._adapter.brightness_tune_custom(self._channel, direction, speed, self._broadcast_mode, self._mode)

    def brightness_tune_step(self, direction: BrightnessDirection, step: int = None):
        return self._adapter.brightness_tune_step(self._channel, direction, step, self._broadcast_mode, self._mode)

    def set_brightness(self, brightness: float) -> [(bool, ModuleInfo)]:
        return self._adapter.set_brightness(self._channel, brightness, self._broadcast_mode, self._mode)


class RGBLed(Dimmer):

    def roll_rgb_color(self) -> [(bool, ModuleInfo)]:
        return self._adapter.roll_rgb_color(self._channel, self._broadcast_mode, self._mode)

    def switch_rgb_color(self) -> [(bool, ModuleInfo)]:
        return self._adapter.switch_rgb_color(self._channel, self._broadcast_mode, self._mode)

    def switch_rgb_mode(self) -> [(bool, ModuleInfo)]:
        return self._adapter.switch_rgb_mode(self._channel, self._broadcast_mode, self._mode)

    def switch_rgb_mode_speed(self) -> [(bool, ModuleInfo)]:
        return self._adapter.switch_rgb_mode_speed(self._channel, self._broadcast_mode, self._mode)

    def set_rgb_brightness(self, red: float, green: float, blue: float) -> [(bool, ModuleInfo)]:
        return self._adapter.set_rgb_brightness(self._channel, red, green, blue, self._broadcast_mode, self._mode)
